#include "tunnel.h"
#include <stdlib.h>
#include "sockio.h"
#include <string.h>
#include <assert.h>
#include <net/if.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <netinet/ip.h>
#include <errno.h>
#include <unistd.h>
#include <linux/if_tunnel.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <asm/types.h>
#include <net/if_arp.h>
#include <time.h>
#include <fcntl.h>
#include <sys/uio.h>
#include <stdio.h>
#include <net/route.h>


int errno;
/*
 * set_tunnel_up
 * dev: name of tunnel
 *
 * Set the tunnel up
 *
 * Returns; 0 if successful, else -1
 */

int set_tunnel_up(const char *dev)
{
        struct ifreq ifr;
        int     do_set;

        do_set=0;
        if(dev==NULL) {
                printf("set_tunnel_up:dev name\n");
                return -1;
        }

        memcpy(ifr.ifr_name, dev, IFNAMSIZ);

        /*get interface flag*/
        if(tunnel_msg(SIOCGIFFLAGS, &ifr) <0)
                return -1;

        if((ifr.ifr_flags & IFF_UP)==0) {
                ifr.ifr_flags |=IFF_UP;
                do_set=1;
        }

        /*set the dev up*/
        if(do_set==1) {
                if(tunnel_msg(SIOCSIFFLAGS, &ifr) <0)
                        return -1;
        }

        return 0;
}

/**
 * add_tunnel_ip:
 * cmd: add interface ip: SIOCSIFADDR
 * dev: name of tunnel
 *
 * Add local ip or netmask on the interface of tunnel
 *
 * Returns: 0 if successful, else -1
 */
int add_tunnel_ip(int cmd,const char *dev, struct in_addr addr)
{
        struct sockaddr_in if_sin;
        struct ifreq       ifr;

        /*cp interface addr to interface request:ifr*/
        if_sin.sin_family=AF_INET;
        memcpy(&(if_sin.sin_addr), &addr,sizeof(addr));
        memcpy(&(ifr.ifr_addr), &if_sin, sizeof(struct sockaddr));
        strncpy(ifr.ifr_name, dev, sizeof(ifr.ifr_name));

        if(tunnel_msg(cmd, &ifr) <0)
                return -1;

        return 0;
}

/*
 * tunnel_msg:
 * cmd: add tunnel: SIOCADDTUNNEL, del tunnel:=SIOCDELTUNNEL
 *      set tunnel: SIOCSIFFLAGS,  add ip: SIOCSIFADDR
 * ifr: interface request
 *
 * Establish tunnel, Set tunnel up, Set tunnel down and Add tunnel IP
 *
 * Returns: 0 if successful, else -1
 */
int tunnel_msg(int cmd,void *ifr)
{
        int s;

        if((s=socket(AF_INET, SOCK_DGRAM, 0)) <0) {
                printf("tunnel_msg: socket error\n");
                return -1;
        }

        if(ioctl(s,cmd,ifr)!=0) {
                printf("ioctl error %s \n",strerror(errno));
                close(s);
                return -1;
        }

        close(s);
        return 0;
}

/**
 * check_tunnel:
 * ifr: interface request
 * dev: interface name
 *
 * Check if the tunnel interface exists
 *
 * Returns: 0 if successful, else -1
 */
int check_tunnel(struct ifreq *ifptr,const char *dev)
{
        int            len;
        struct ifconf ifc;
        #define MAX_INTERFACES 50
        char           *ptr,buffer[MAX_INTERFACES * sizeof(struct ifreq)];
        struct ifreq   *ifr;

        len=sizeof(struct sockaddr);
        ifc.ifc_len=sizeof(buffer);
        ifc.ifc_buf=buffer;

        if(tunnel_msg(SIOCGIFCONF,&ifc)<0)
                return -1;

        /*
         *check if dev exists
         */
        for (ptr=buffer; ptr <buffer+ifc.ifc_len;) {
                ifr=(struct ifreq *) ptr;
                if(strncmp(dev, ifr->ifr_name, 20) == 0) {
                        memcpy(ifptr,ifr,sizeof(struct ifreq));
                        return 0;
                }

                ptr +=sizeof(ifr->ifr_name)+len;

        }

        return -1;
}

/*
 * init_tunnel_param:
 * tun_param: ip tunnel param
 * dev: the name of tunnel
 * remote: IP address of remote end of tunnel
 * local: IP address of local end of tunnel
 *
 * Init the tunnel parameter
 *
 * Returns: 0 if successful, else -1
 */
int init_tunnel_param(struct ip_tunnel_parm *tun_param,const char *dev,int ttl,
                                        __u32 remote, __u32 local)
{
        if(tun_param==NULL) {
                printf("Tunnel Parameter does not exist\n");
                return -1;
        }

        memset(tun_param, 0, sizeof(tun_param));
        tun_param->iph.version=4;
        tun_param->iph.ihl=5;
        tun_param->iph.frag_off=htons(IP_DF);           /*Flag: Do not fragment*/
        tun_param->iph.ttl=ttl;
        tun_param->iph.protocol=IPPROTO_IPIP;           /*Protocl: IPIP*/

        strncpy(tun_param->name,dev,IFNAMSIZ);

        /*remote and local is used for add tunnel */
        tun_param->iph.daddr=remote;
        tun_param->iph.saddr=local;

        return 0;
}

/**
 * rt_add:
 * dst: mn address
 * gateway: coa address
 * dev: device name
 *
 * Add tunnel route in route table
 *
 * Returns: 0 if successfull, else -1
 */
int rt_add(struct in_addr dst, char *dev)
{
        struct rtentry rt;
        struct sockaddr_in sa;

        memset(&rt, 0, sizeof(rt));
        memset(&sa, 0, sizeof(sa));
        sa.sin_family=AF_INET;
        sa.sin_addr=dst;

        /*add mn address in the route table*/
        rt.rt_flags=RTF_UP|RTF_HOST;
        rt.rt_dev=dev;
        memcpy(&rt.rt_dst,&sa,sizeof(sa));
        if(tunnel_msg(SIOCADDRT,&rt)<0)
                return -1;

        return 0;
}

/**
 *add_tunnel:
 *dev: name of tunnel
 *remote: address of remote end of tunnel
 *local: address of local end of tunnel
 *
 * Add the tunnel and set tunnel up
 *
 * Returns: 0 if successful, else -1
 */
int add_tunnel( const char *dev,struct in_addr remote, struct in_addr local)
{
        static struct ip_tunnel_parm tun_param;
        struct ifreq          ifr,ifrtmp;
        int                   re,do_set;
        int                   cmd;
        char                  str[128];
        struct  sockaddr      *sa;
        struct  sockaddr_in   *sin;

        cmd=SIOCADDTUNNEL;
        do_set=0;
        if(dev ==NULL) {
                printf("add_tunnel: imput parameter error\n");
                return -1;
        }

        /*check if the tunnel exists*/
        if(check_tunnel(&ifrtmp,dev)==0) {
                if((sa=&ifrtmp.ifr_addr)==NULL)
                        return -1;            
	  do_set=1;
        }

        /*init the tunnel param*/
        if(do_set==0) {
                re=init_tunnel_param(&tun_param,dev,64,remote.s_addr,local.s_addr);
                if(re < 0)
                        return -1;

                ifr.ifr_ifru.ifru_data=(void *)&tun_param;
                strncpy(ifr.ifr_name, "tunl0", IFNAMSIZ);

                if(tunnel_msg(cmd,&ifr) <0)
                return -1;

        }

	 if(add_tunnel_ip(SIOCSIFADDR,dev,local) < 0)
                return -1;

        if(set_tunnel_up(dev) < 0)
                return -1;

        return 0;
}

void mn_route_add_tunnel(struct in_addr cli,struct in_addr ha,struct in_addr dest,char *dev,char *head)
{
  int err;
  struct in_addr ha_addr,cli_addr;
	
  ha_addr  = ha;
  cli_addr = cli;
    
  err = add_tunnel(dev,ha_addr,cli_addr);

  if(err < 0)
      printf("add tunnel: %s fail\n", dev);

  err=rt_add(dest, dev);
  if(err < 0)
     printf("add mn  route fail \n"); 
  
}

main(int argc,char *argv[])
{
  struct in_addr client,ha,dest;
  char *dev;

  printf("%d \n",argc);

  if(argc!=5) {
   printf("tunnel_test client ha dest dev\n");
   return -1;
  }


  memset((char *) &ha,0, sizeof(ha));
  inet_aton((char *)argv[1], &ha);
  memset((char *) &client, 0, sizeof(client));
  inet_aton((char *) argv[2], &client);
  memset((char *) &dest, 0, sizeof(dest));
  inet_aton((char *) argv[3], &dest);
  memcpy(dev,argv[4],IFNAMSIZ);

  mn_route_add_tunnel(client,ha,dest,dev,NULL);
}
